/* File: process.c
 * Desc: AFP request processor
 * Auth: San Mehat
 * Note:
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <znodelist.h>
#include <afp.h>
#include <config.h>

extern CONFIGURATION *Tconfigp_ptr;

/* Determine what the other guy wants... do it... */
int 
process_packet (void *pkt, Z_NODE * node)
{
	AFP_HDR *hdr = (AFP_HDR *) pkt;

#if 0
	printf ("[%s] Processing packet from node %d (txid=%d)\n", mytime (), node->I_nodenum, hdr->UI_txid);
#endif

	switch (hdr->UI_cmd)
	{
	case REQ_CONFIG:
		{
			printf ("[%s] Node %d requested configuration data\n", mytime (), node->I_nodenum);

			send_config (node, pkt);	// XXX check return code and queue a retry if failure
#if 0
			printf ("[%s] debug: REQ_CONFIG Phase transition from %d to %d\n",
							mytime (),
							node->phase.I_operation,
							PHASE_CONFIG);
#endif
			node->phase.I_operation = PHASE_CONFIG;
			node->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_CONFIG_TIMEOUT);
			break;
		}

	case ACK_RESET:
		printf ("[%s] Node %d acknowledged reset request\n", mytime (), node->I_nodenum);
		break;

	case ACK_STATUS:
		{
			Z_STATS *statsptr;

			/* Copy the status information */
			statsptr = (Z_STATS *) & ((char *) pkt)[sizeof (AFP_HDR)];
			memcpy (&node->ZS_stats, statsptr, sizeof (Z_STATS));
			printf ("[%s] Node %d reported memf=%d memt=%d temp=%d load=%d\n", mytime (), node->I_nodenum,
							node->ZS_stats.UI_memfree, node->ZS_stats.UI_memtotal,
							node->ZS_stats.UI_temperature, node->ZS_stats.UI_cpuload);
#if 0
			printf ("[%s] debug: ACK_STATUS Phase transition from %d to %d\n",
							mytime (),
							node->phase.I_operation,
							PHASE_IDLE);
#endif
			node->phase.I_operation = PHASE_IDLE;
			node->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_IDLE_TIMEOUT);
			break;
		}

	case SIG_READY:
		printf ("[%s] Node %d signals ready to begin processing (IP_PORTFW ON)\n", mytime (), node->I_nodenum);
		node->flags = FLAG_POWR | FLAG_INST;	/* XXX: POWR and INST should
																					   not be here */
		node->UI_failcount = 0;			/* Reset error counter... he's done his time.. paid his dues */
		node->UI_afptstamp = 0;
		/*ip_portfw_add (node);TEMP*/
#if 0
		printf ("[%s] debug: SIG_READY: Phase transition from %d to %d\n",
						mytime (),
						node->phase.I_operation,
						PHASE_IDLE);
#endif
		node->phase.I_operation = PHASE_IDLE;
		node->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_IDLE_TIMEOUT);
		break;

	default:
		{
			struct in_addr addr;

			printf ("[%s] Node %d command error (%d)\n", mytime (), node->I_nodenum, hdr->UI_cmd);
			addr.s_addr = node->UI_ipaddr;
			send_ll_nak (addr);
		}
	}
	return (0);
}

int 
send_data (Z_NODE * node, char *data, unsigned int UI_size)
{
	int I_sockfd;
	struct sockaddr_in sock_addr;
	struct in_addr debug;

	bzero ((char *) &sock_addr, sizeof (sock_addr));
	sock_addr.sin_family = AF_INET;
	sock_addr.sin_port = htons (6443);
	sock_addr.sin_addr.s_addr = ((Tconfigp_ptr->UI_zibaddr & 0x00ffffff) | (node->I_nodenum << 24));

	debug.s_addr = sock_addr.sin_addr.s_addr;
// printf("addr is %s\n",inet_ntoa(debug));
	if ((I_sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
	{
		perror ("socket");
		printf ("[send_data] Unable to send data to slave at %s (%d)\n",
						inet_ntoa (sock_addr.sin_addr),
						I_sockfd);
		return (-1);
	}
	//printf("[send_data] Sending data packet 0x%x of size %d\n",data,UI_size);
	if (sendto (I_sockfd, (void *) data, UI_size, 0, (struct sockaddr *) &sock_addr, sizeof (sock_addr)) < 0)
	{
		perror ("sendto");
		printf ("[send_data] Unable to send data to slave at %s (%d)\n",
						inet_ntoa (sock_addr.sin_addr));
		return (-1);
	}
	close (I_sockfd);
	return (0);
}

int 
send_config (Z_NODE * node, void *req)
{
	char *pkt;
	AFP_HDR *reqhdr = (AFP_HDR *) req;
	AFP_HDR *pkthdr;
	CONFIG_PARTITION *Tpartitionp_ptr;
	CONFIG_NODELIST *Tconfignodelistp_ptr;
	AFP_CONFIG *Tafpconfigp_ptr;
	int I_numentries = 0;

	/* 
	   Build up our configuration packet.. and send it to the node
	 */
	pkthdr = (AFP_HDR *) pkt = (char *) malloc (4096);
	memset (pkthdr, 0, 4096);

	pkthdr->UI_txid = reqhdr->UI_txid;
	pkthdr->UI_cmd = ACK_CONFIG;
	/* 
	   Configuration packet contains a list of partitions this slave is to run.
	   This consists of the portbindings and docroots for a server
	 */

	/* 
	   Walk each partition and see which ones this node belongs to.
	   For each partition it's a member of, add to the packet
	 */
	Tpartitionp_ptr = Tconfigp_ptr->Tpartition_root;
	while (Tpartitionp_ptr)
	{
		Tconfignodelistp_ptr = Tpartitionp_ptr->Tnodelist_root;
		while (Tconfignodelistp_ptr)
		{
			if (Tconfignodelistp_ptr->I_nodenumber == node->I_nodenum)
			{
				int I_c1;

				/*
				   He's a member.... tell him about it.
				   First increase the size of our packet.. and increment data length;
				 */
				I_numentries++;

				/* 
				   Update datalength
				 */
				pkthdr->UI_dlen = (sizeof (AFP_CONFIG) * I_numentries);

				/*
				   Position pointer at first.
				 */
				Tafpconfigp_ptr = (AFP_CONFIG *) (pkt + sizeof (AFP_HDR));
				/*
				   now Fast forward to end
				 */
				for (I_c1 = 1; I_c1 < I_numentries; I_c1++, Tafpconfigp_ptr++)
				{
				}
				strcpy (Tafpconfigp_ptr->SZ_docroot, Tpartitionp_ptr->SZ_documentpath);
				Tafpconfigp_ptr->US_portbinding = Tpartitionp_ptr->US_portbinding;
			}
			Tconfignodelistp_ptr = Tconfignodelistp_ptr->next;
		}
		Tpartitionp_ptr = Tpartitionp_ptr->next;
	}
	/*
	   Now send the data
	 */
	if (send_data (node, pkt, (sizeof (AFP_HDR) + (sizeof (AFP_CONFIG) * I_numentries))) < 0)
	{
		printf ("[send_config] data send failed\n");
		free (pkt);
		return (-1);
	}
	free (pkt);
	node->phase.I_operation = PHASE_CONFIG;
	node->phase.UI_alarmtime = (((unsigned int) time (NULL)) + PHASE_CONFIG_TIMEOUT);
	return (0);
}
