/* File: portfw.h
 * Desc: Definitions for communicating with IP Port Forwarder
 * Auth: Matthew Kellett
 * Note: Adapted from ip_fw.h  Relevant changes should be reflected here
*/

#define IP_FW_BASE_CTL  	64	/* base for firewall socket options */

#define IP_FW_SHIFT		8	/* shift count for type (chain) */

#define IP_FW_PORTFW		6

#define IP_FW_INSERT		(IP_FW_BASE_CTL)
#define IP_FW_APPEND		(IP_FW_BASE_CTL+1)
#define IP_FW_DELETE		(IP_FW_BASE_CTL+2)
#define IP_FW_FLUSH			(IP_FW_BASE_CTL+3)
#define IP_FW_ZERO			(IP_FW_BASE_CTL+4)
#define IP_FW_POLICY		(IP_FW_BASE_CTL+5)
#define IP_FW_CHECK			(IP_FW_BASE_CTL+6)
#define IP_FW_MASQ_TIMEOUTS	(IP_FW_BASE_CTL+7)


#define IP_PORTFW_ADD		(IP_FW_APPEND | (IP_FW_PORTFW << IP_FW_SHIFT))
#define IP_PORTFW_DEL		(IP_FW_DELETE | (IP_FW_PORTFW << IP_FW_SHIFT))
#define IP_PORTFW_FLUSH  	(IP_FW_FLUSH  | (IP_FW_PORTFW << IP_FW_SHIFT))
#define IP_PORTFW_BIND	 	(IP_FW_POLICY | (IP_FW_PORTFW << IP_FW_SHIFT))
#define IP_PORTFW_STATS		(IP_FW_CHECK  | (IP_FW_PORTFW << IP_FW_SHIFT))

#define IP_PORTFW_LBM_NAME_LEN  32 

#ifndef __u16
#define __u16 unsigned short
#endif
#ifndef __u32
#define __u32 unsigned long
#endif

struct ip_portfw_edits {
       __u16           protocol;       /* Which protocol are we talking? */
       __u32           laddr, raddr;   /* Local and remote address */
       __u16           lport, rport;   /* Local and remote port */
		short 			weight;			/* Weight for this server */
		int				memory;			/* Memory used on real server */
		int 			load;			/* load avg on machine */
		char			lbm_name[IP_PORTFW_LBM_NAME_LEN];	/* name of module for ip_portfw */
};
